"""This module contains the metabolic model preparation tools"""
# print('This is local package') # uncomment this then upload to pypi

from .core.component import get_links_component, get_component, get_gene_id
from .basic_model import BasicModel
